#include <QtGui>
#include <QtOpenGL/QGLWidget>

#include "glarea.h"

GLArea::GLArea (QWidget * parent) :QGLWidget (parent)
{
	eyeDist = 15.0;
	eyePhi = 0.0;
	eyeTheta = 0.0;

	showWireframed = false;
	showSeamlessTextureU = showSeamlessTextureV =false;
	showBilinear = true;
	showMipmap = true;
}

void GLArea::setTextureOptions(){
	if (showMipmap && showBilinear)  {
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	}
	if (showMipmap && !showBilinear)  {
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	}
	if (!showMipmap && showBilinear)  {
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	}
	if (!showMipmap && !showBilinear)  {
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
		glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	}
}

void GLArea::setMipmap(bool mode){
	showMipmap = mode;
	setTextureOptions();
	update();
}

void GLArea::setBilinear(bool mode){
	showBilinear = mode;
	setTextureOptions();
	update();
}

void GLArea::setTextureNumber(int textureNumber){
	makeCurrent();
	bindTexture(
		QString("%1.dds").arg(textureNumber,1,16)
	);
	setTextureOptions();
	update();
}

void GLArea::setTexturedOrWireframe(bool mode){
	showWireframed = !mode;
	update();
}

void GLArea::setSeamlessTextureU(bool mode){
	showSeamlessTextureU = mode;
	update();
}

void GLArea::setSeamlessTextureV(bool mode){
	showSeamlessTextureV = mode;
	update();
}


/* procedural generation of shapes (as vertex arrays) */

// helper func: returns a random number in -1 +1
static float rand11(){
	return (rand()%1000/500.0f)-1;
}

// helper func: add noise to a float
static void noise(float &f, float n){
	f +=  ( rand11()+rand11()+rand11() )/6.0*n;
}

// helper func: a bit of pseudorandom spatial distort (for the cyl)
static void distort(float &x, float &y, float &z){
	float r = (1.0+0.2*sin(y*3) + 0.01*cos(x*z));
	x*=r;
	z*=r;

	float d = (0.2*sin(y*1.4+5.0));
	x+=d;
	z+=d;
}

// procedural generation of a torus
void GLArea::makeTorus(){
	const int QUAD_X = 40;
	const int QUAD_Y = 20;


	const int N_VERTS =QUAD_X*QUAD_Y;

	const int N_FACES =QUAD_X*QUAD_Y*2; // two tris per face

	static GLfloat vertices[ N_VERTS*3 ];
	static GLfloat normals[ N_VERTS*3 ];
	static GLfloat textCoords[ N_VERTS*2 ];

	float R0 = 0.6*1.1;
	float R1 = 1.05*1.1;
	for (int v=0,i=0,j=0,k=0; v<QUAD_Y; v++)
	for (int u=0; u<QUAD_X; u++)
	{
		float vf = float(v)/QUAD_Y  ;
		float uf = float(u)/QUAD_X ;

		noise(uf,1.0/QUAD_X);
		noise(vf,1.0/QUAD_Y);

		float x0 = (float)sin(vf*M_PI*2);
		float y0 = (float)cos(vf*M_PI*2);

		float x1 = (float)sin(uf*M_PI*2);
		float y1 = (float)cos(uf*M_PI*2);

		vertices[i++] = (x0*R0+R1)*y1;
		vertices[i++] = y0*R0;
		vertices[i++] = (x0*R0+R1)*x1;

		normals[j++] = x0*y1;
		normals[j++] = y0;
		normals[j++] = x0*x1;

		textCoords[k++] = uf*2;
		textCoords[k++] = vf;
	}

	glVertexPointer(3, GL_FLOAT, 0, vertices);
	glNormalPointer( GL_FLOAT, 0, normals);
	glTexCoordPointer( 2, GL_FLOAT, 0, textCoords );

	indices.resize( N_FACES*3 );

	for (int v=0,i=0; v<QUAD_Y; v++)
	for (int u=0; u<QUAD_X; u++)
	{

		int randSplit = rand()%2; // a random diagonal split

		if(randSplit) {
			// first triangle
			indices[i++] = (u+0)%QUAD_X+(v+0)%QUAD_Y*QUAD_X;
			indices[i++] = (u+0)%QUAD_X+(v+1)%QUAD_Y*QUAD_X;
			indices[i++] = (u+1)%QUAD_X+(v+1)%QUAD_Y*QUAD_X;

			// second trianlge
			indices[i++] = (u+1)%QUAD_X+(v+1)%QUAD_Y*QUAD_X;
			indices[i++] = (u+1)%QUAD_X+(v+0)%QUAD_Y*QUAD_X;
			indices[i++] = (u+0)%QUAD_X+(v+0)%QUAD_Y*QUAD_X;
		} else{
			// first triangle
			indices[i++] = (u+0)%QUAD_X+(v+1)%QUAD_Y*QUAD_X;
			indices[i++] = (u+1)%QUAD_X+(v+1)%QUAD_Y*QUAD_X;
			indices[i++] = (u+1)%QUAD_X+(v+0)%QUAD_Y*QUAD_X;

			// second trianlge
			indices[i++] = (u+1)%QUAD_X+(v+0)%QUAD_Y*QUAD_X;
			indices[i++] = (u+0)%QUAD_X+(v+0)%QUAD_Y*QUAD_X;
			indices[i++] = (u+0)%QUAD_X+(v+1)%QUAD_Y*QUAD_X;
		}
	}

	update();

}

// procedural geneartion of a trinagulated cylinder
void GLArea::makeCylinder(){

	const int QUAD_X = 30;
	const int QUAD_Y = 15;


	const int N_VERTS =QUAD_X*(QUAD_Y+1);

	const int N_FACES =QUAD_X*QUAD_Y*2; // two tris per face

	static GLfloat vertices[ N_VERTS*3 ];
	static GLfloat normals[ N_VERTS*3 ];
	static GLfloat textCoords[ N_VERTS*2 ];

	float R = 0.8; // radius
	float H = 3.5; // height
	for (int v=0,i=0,j=0,k=0; v<QUAD_Y+1; v++)
	for (int u=0; u<QUAD_X; u++)
	{
		float vf = float(v)/QUAD_Y;
		float uf = float(u)/QUAD_X;// - vf*0.25;

		noise(uf,1.0/QUAD_X);
		if (v!=0 && v!=QUAD_Y) noise(vf,1.0/QUAD_Y);  // distort parametric pos uv

		// from uv to xyz
		float x = (float)sin(uf*M_PI*2);
		float y = (vf-0.5)*H;
		float z = (float)cos(uf*M_PI*2);

		normals[j++] = x*1.2;
		normals[j++] = 0;
		normals[j++] = z*1.2;

		textCoords[k++] = uf*2;
		textCoords[k++] = 1-vf;

		distort(x,y,z); // distort position

		vertices[i++] = x*R;
		vertices[i++] = y;
		vertices[i++] = z*R;

	}

	glVertexPointer(3, GL_FLOAT, 0, vertices);
	glNormalPointer( GL_FLOAT, 0, normals);
	glTexCoordPointer( 2, GL_FLOAT, 0, textCoords );

	indices.resize( N_FACES*3 );

	for (int v=0,i=0; v<QUAD_Y; v++)
	for (int u=0; u<QUAD_X; u++)
	{
		// first triangle forming the quad
		indices[i++] = (u+0)%QUAD_X+(v+0)%(QUAD_Y+1)*QUAD_X;
		indices[i++] = (u+1)%QUAD_X+(v+0)%(QUAD_Y+1)*QUAD_X;
		indices[i++] = (u+1)%QUAD_X+(v+1)%(QUAD_Y+1)*QUAD_X;

		// second tri forming the quad
		indices[i++] = (u+1)%QUAD_X+(v+1)%(QUAD_Y+1)*QUAD_X;
		indices[i++] = (u+0)%QUAD_X+(v+1)%(QUAD_Y+1)*QUAD_X;
		indices[i++] = (u+0)%QUAD_X+(v+0)%(QUAD_Y+1)*QUAD_X;
	}

	update();

}

/* creation of shaders code in GLSL (as strings) */

// code in GLSL for declarations (common to frag and vert shader)
QString GLArea::getVaryingsDeclarations(bool seamlessU, bool seamlessV, bool useTexture) {

	// for all...
	QString res = QString(
		//"#version 130\n"
		"varying float diffuse;\n"
	);

	if (seamlessU && seamlessV) res+=QString(
		"varying vec2 uv0; /* one way */ \n"
		"varying vec2 uv1; /* other way */ \n"
	);
	else if (seamlessU) res+=QString(
		"varying float v; /* text coord v */ \n"
		"varying float u0; /* u: one way */ \n"
		"varying float u1; /* u: other way */ \n"
	);
	else if (seamlessV) res+=QString(
		"varying float u; /* text coord u */ \n"
		"varying float v0; /* v: one way */ \n"
		"varying float v1; /* v: other way */ \n"
	);
	else if (useTexture) res+=QString(
		"varying vec2 uv; /* text coords */ \n"
	);

	res+="\n";
	return res;
}

// code in GLSL for Vertex Shaders
QString GLArea::getVertexShader(bool seamlessU, bool seamlessV, bool useTexture) {

	QString res = getVaryingsDeclarations( seamlessU, seamlessV, useTexture );

	res += QString(
		"void main(){\n"
		"\n"
		"  /* plain transform and lighting... */\n"
		"  gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
		"  diffuse = ( gl_NormalMatrix * gl_Normal ).z ;\n"
		"  diffuse = ((diffuse<0.0)?-0.3*diffuse:diffuse ); /* back light */\n"
		"  diffuse = diffuse * 0.6+0.6;  /* ambient */ \n"
	);

	if (!useTexture) res+= QString(
		"  diffuse *= gl_Color.x;\n"
	); else {
		res+= QString(
		"\n"
		"  /* pass down texture coordiantes */ \n"
		);
	}

	if (seamlessU && seamlessV) res+=QString(
		"  uv0 = fract( gl_MultiTexCoord0.st ); \n"
		"  uv1 = fract( gl_MultiTexCoord0.st + vec2(0.5,0.5) ) - vec2(0.5,0.5); /* see paper */ \n"
	);
	else if (seamlessU) res+=QString(
		"  v  = gl_MultiTexCoord0.y ; \n"
		"  u0 = fract( gl_MultiTexCoord0.x ) ; \n"
		"  u1 = fract( gl_MultiTexCoord0.x + 0.5 ) - 0.5; /* see paper */ \n"
	);
	else if (seamlessV) res+=QString(
		"  u  = gl_MultiTexCoord0.x ; \n"
		"  v0 = fract( gl_MultiTexCoord0.y ) ; \n"
		"  v1 = fract( gl_MultiTexCoord0.y + 0.5 ) - 0.5; /* see paper */ \n"
	);
	else if (useTexture) res+=QString(
		"  uv = gl_MultiTexCoord0.st ; \n"
	);

	res += QString(
		"\n"
		"}\n"
	);

	return res;
}

// code in GLSL for fragment shaders
QString GLArea::getFragmentShader(bool seamlessU, bool seamlessV, bool useTexture) {
	QString res = getVaryingsDeclarations( seamlessU, seamlessV, useTexture );

	if (useTexture) res += QString(
		"uniform sampler2D mySampler;\n\n"
	);

	res += QString(
		"void main(){\n"
	);
	if (!useTexture) res += QString(
		"  gl_FragColor.rgb = vec3( diffuse ); \n"
	);
	else if (!seamlessU && !seamlessV) res += QString(
		"  gl_FragColor.rgb = texture2D( mySampler, uv ).rgb * diffuse; \n"
	);
	else {
		res += QString(
			"  vec2 uvT; \n"
		);
		if (seamlessU && seamlessV) res += QString(
			"  uvT.x = ( fwidth( uv0.x ) < fwidth( uv1.x )-0.001 )? uv0.x : uv1.x ; \n"
			"  uvT.y = ( fwidth( uv0.y ) < fwidth( uv1.y )-0.001 )? uv0.y : uv1.y ; \n"
		);
		else if (seamlessU) res += QString(
			"  uvT.x = ( fwidth( u0 ) < fwidth( u1 )-0.001 )? u0 : u1 ; \n"
			"  uvT.y = v; \n"
		);
		else if (seamlessV) res += QString(
			"  uvT.x = u ; \n"
			"  uvT.y = ( fwidth( v0 ) < fwidth( v1 )-0.001 )? v0 : v1; \n"
		);

		res += QString(
			"  gl_FragColor.rgb = texture2D( mySampler, uvT  ).rgb * diffuse; \n"
		);
	}

	res += QString(
		"  gl_FragColor.a = 1.0; \n"
		"}\n"
	);

	return res;

}

void GLArea::makeShaders(){

	// progPlain: used for wireframes (no texture)
	progPlain.removeAllShaders();
	progPlain.addShaderFromSourceCode( QGLShader::Vertex,   getVertexShader  (0,0,0));
	progPlain.addShaderFromSourceCode( QGLShader::Fragment, getFragmentShader(0,0,0));
	progPlain.link();

	// progSeamless: all combinations of seamless or not on U and V
	for (int j=0; j<2; j++)
	for (int i=0; i<2; i++) {
		progSeamless[i][j].removeAllShaders();
		progSeamless[i][j].addShaderFromSourceCode( QGLShader::Vertex,  getVertexShader(i,j,1) );
		progSeamless[i][j].addShaderFromSourceCode( QGLShader::Fragment,getFragmentShader(i,j,1) );
		progSeamless[i][j].link();
		progSeamless[i][j].setUniformValue("mySampler",0);
	}

}

QString GLArea::getCurrentVertexShader() const{
	return getVertexShader(showSeamlessTextureU, showSeamlessTextureV, true);
}

QString GLArea::getCurrentFragmentShader() const{
	return getFragmentShader(showSeamlessTextureU, showSeamlessTextureV, true);
}

void GLArea::initializeGL()
{
	glClearColor(1, 1, 1, 0);
  glEnable(GL_DEPTH_TEST);
	glPolygonOffset(1,1);
	setTextureNumber( 0 );
	makeShaders();
	makeCylinder();
}

void GLArea::paintGL()
{

  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  int w = width(), h = height();
  glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	gluPerspective(30, w/(float)h, 0.1, 100);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
	glTranslatef(0,0,-eyeDist);
	glRotatef(eyeTheta, 1,0,0);
	glRotatef(eyePhi, 0,1,0);
	glScalef(2,2,2);

	/* draw geometry (vertex arrays) */

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	if (showWireframed) {

		progPlain.bind();

		/* show wireframed (first pass) */
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		glColor3f(0.7,0.7,0.7); // lines are dark
		glDisable(GL_POLYGON_OFFSET_FILL);
		glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, &(indices[0]) );

		/* (second pass) */
		glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
		glColor3f(1,1,1);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, &(indices[0]) );

	} else {

		/* show textured, using appropriate shader program */
		progSeamless[showSeamlessTextureU][showSeamlessTextureV].bind();

		glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, &(indices[0]) );
	}

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);

} 

void GLArea::mousePressEvent(QMouseEvent * e)
{
	lastMousePos = e->pos();
	e->accept();
	setFocus();
	updateGL();
}

void GLArea::mouseMoveEvent(QMouseEvent * e)
{
	QPointF dp = e->pos() - lastMousePos;
	lastMousePos = e->pos();

	eyePhi += dp.x();
	eyeTheta += dp.y();
	if (eyeTheta>+90) eyeTheta=+90;
	if (eyeTheta<-90) eyeTheta=-90;
	updateGL();
}

void GLArea::wheelEvent(QWheelEvent * e)
{
	/* wheel ==> zoom */
	if (e->delta()>0) eyeDist *= 1.1;
	else eyeDist /= 1.1;
	if (eyeDist>40) eyeDist=40;
	if (eyeDist<3) eyeDist=3;
	updateGL();
}
